"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


import sys

from asyncio import run

from expedite.client.room import oper

from click import IntRange, option, version_option, group, Path

from expedite import __versdata__
from expedite.config import standard
from expedite.client.meet import talk
from expedite.client.base import find_name, find_size, bite_file
from expedite.client.util import facade_exit
from websockets.exceptions import InvalidURI


def work() -> None:
    talk()
    try:
        run(oper())
    except OSError as expt:
        run(facade_exit(None, False, "oser"))
        sys.exit(standard.client_exit)
    except InvalidURI as expt:
        run(facade_exit(None, False, "iuri"))
        sys.exit(standard.client_exit)
    except KeyboardInterrupt as expt:
        run(facade_exit(None, False, "intr"))
        sys.exit(standard.client_exit)


@group(
    name="expedite",
    context_settings={"show_default": True},
)
@option(
    "-a",
    "--addr",
    "addr",
    type=str,
    default=standard.client_addr,
    required=True,
    help="Set the address for the service endpoint"
)
@option(
    "-t",
    "--time",
    "time",
    type=IntRange(5, 30),
    default=standard.client_time,
    required=False,
    help="Set the expiry period for participants"
)
@option(
    "-e",
    "--endo",
    "endo",
    type=str,
    default=standard.client_endo,
    required=False,
    help="Set the identity of the opposing client"
)
@version_option(
    version=__versdata__, prog_name="Expedite Client by Akashdeep Dhar"
)
def main(
    addr: str = standard.client_addr,
    time: int = standard.client_time,
    endo: str = standard.client_endo,
):
    standard.client_addr = addr
    standard.client_time = time
    standard.client_endo = endo


@main.command(
    name="send",
    help="Deliver file through an encrypted transfer",
    context_settings={"show_default": True},
)
@option(
    "-p",
    "--pswd",
    "pswd",
    type=str,
    default=standard.client_pswd,
    help="Set the password for delivering encryption",
)
@option(
    "-f",
    "--file",
    "file",
    type=Path(exists=True),
    required=True,
    help="Set the filepath for delivering to network",
)
def send(
    pswd: str = standard.client_pswd,
    file: str = standard.client_file
) -> None:
    standard.client_pswd = pswd
    standard.client_file = file
    standard.client_filesize = find_size()
    standard.client_filename = find_name()
    standard.client_bind = bite_file()
    standard.client_plan = "SEND"
    work()


@main.command(
    name="recv",
    help="Collect file through an encrypted transfer",
    context_settings={"show_default": True},
)
@option(
    "-p",
    "--pswd",
    "pswd",
    type=str,
    required=True,
    help="Set the password for collecting encryption"
)
def recv(
    pswd: str = standard.client_pswd
) -> None:
    standard.client_pswd = pswd
    standard.client_plan = "RECV"
    work()



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


from asyncio import get_event_loop

from expedite.server.room import oper

from click import IntRange, option, version_option, command

from expedite import __versdata__
from expedite.config import standard
from expedite.server.meet import talk
from expedite.view import failure

import sys

from websockets import serve


def work():
    func = serve(oper, standard.server_addr, standard.server_port)
    get_event_loop().run_until_complete(func)
    get_event_loop().run_forever()


@command(
    name="expedite",
    context_settings={"show_default": True},
)
@option(
    "-a",
    "--addr",
    "addr",
    type=str,
    default=standard.server_addr,
    required=False,
    help="Set the interface for the service endpoint"
)
@option(
    "-p",
    "--port",
    "port",
    type=IntRange(min=64, max=65535),
    default=standard.server_port,
    required=False,
    help="Set the port value for the service endpoint"
)
@version_option(
    version=__versdata__, prog_name="Expedite Server by Akashdeep Dhar"
)
def main(addr: str = standard.server_addr, port: int = standard.server_port) -> None:
    try:
        standard.server_addr = addr
        standard.server_port = port
        talk()
        work()
    except KeyboardInterrupt:
        failure("Exiting.")
        sys.exit(1)



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


import asyncio
import time
from json import dumps
from expedite.view import warning, general
from expedite.config import standard
from websockets.legacy.client import WebSocketClientProtocol
from expedite.client.base import ease_size, read_file, fuse_file
from hashlib import sha256
from expedite.client.util import facade_exit
from tqdm.asyncio import trange
from tqdm.contrib.logging import logging_redirect_tqdm
from tqdm.asyncio import tqdm


async def deliver_connection_to_server(sock: WebSocketClientProtocol) -> bool:
    general("Attempting to connect to the network.")
    try:
        await sock.send(dumps({"call": "join", "plan": standard.client_plan, "scan": standard.client_endo, "wait": standard.client_time}))
        return True
    except Exception as expt:
        warning(f"Failed to connect to the network - {expt}.")
        return False


async def collect_permission_to_join(iden: str = standard.client_iden) -> bool:
    standard.client_iden = iden
    general(f"Successfully connected to the network.")
    warning(f"You are now identified as {iden} in the network.")
    return True


async def deliver_suspension_from_expiry(sock: WebSocketClientProtocol) -> None | bool:
    if not standard.client_pair:
        general("Attempting to abandon from the network after expiry.")
        await sock.send(dumps({"call": "rest"}))
        await facade_exit(sock, False, "rest")
    else:
        return False


async def collect_connection_from_pairness(iden: str = standard.client_endo) -> bool:
    standard.client_endo = iden
    standard.client_pair = True
    general(f"Attempting pairing with {standard.client_endo}.")
    warning(f"Starting transmission.")
    return True


async def collect_metadata(name: str = standard.client_filename, size: str = standard.client_filesize, chks: int = standard.client_chks) -> bool:
    standard.client_filename, standard.client_filesize, standard.client_chks = name, size, chks
    general(f"Collecting metadata for '{standard.client_filename}' ({ease_size(standard.client_filesize)}) from {standard.client_endo}.")
    return True


async def deliver_metadata(sock: WebSocketClientProtocol):
    await sock.send(dumps({"call": "meta", "name": standard.client_filename, "size": standard.client_filesize, "chks": len(standard.client_bind)-1}))
    general(f"Delivering metadata for '{standard.client_filename}' ({ease_size(standard.client_filesize)}) to {standard.client_endo}.")
    return True


async def deliver_dropping_summon(sock: WebSocketClientProtocol) -> bool:
    await sock.send(dumps({"call": "drop"}))
    general(f"Delivering collection summon to {standard.client_endo}.")
    return True


async def collect_dropping_summon() -> bool:
    general(f"Collecting delivering summon from {standard.client_endo}.")
    return True


async def deliver_contents(sock: WebSocketClientProtocol) -> bool:
    with logging_redirect_tqdm():
        with tqdm(total=standard.client_filesize, unit="B", unit_scale=True, unit_divisor=1024, leave=False, initial=0) as prog:
            for indx in range(0, len(standard.client_bind) - 1):
                bite = read_file(standard.client_bind[indx], standard.client_bind[indx + 1])
                # general(f"Delivering file contents to {standard.client_endo} (SHA256 {sha256(bite).hexdigest()}).")
                prog.update(standard.client_bind[indx + 1] - standard.client_bind[indx])
                await sock.send(bite)
                await asyncio.sleep(0)
        return True


async def collect_contents(sock: WebSocketClientProtocol, pack: bytes = b"") -> bool:
    fuse_file(pack)
    with logging_redirect_tqdm():
        with tqdm(total=standard.client_filesize, unit="B", unit_scale=True, unit_divisor=1024, leave=False, initial=len(pack)) as prog:
            for indx in range(standard.client_chks - 1):
                mesgcont = await sock.recv()
                if isinstance(mesgcont, bytes):
                    fuse_file(mesgcont)
                    prog.update(len(mesgcont))
                    await asyncio.sleep(0)
    # general(f"Collecting file contents from {standard.client_endo} - (SHA256 {sha256(pack).hexdigest()})")
    return True


async def deliver_digest_checks(sock: WebSocketClientProtocol) -> bool:
    general(f"Delivering contents digest for confirmation.")
    await sock.send(dumps({"call": "hash", "data": standard.client_hash.hexdigest()}))
    return True


async def collect_digest_checks() -> bool:
    general(f"Collecting contents digest for confirmation.")
    return True


async def deliver_confirmation(sock: WebSocketClientProtocol, data: str = standard.client_hash.hexdigest()) -> bool:
    if data == standard.client_hash.hexdigest():
        general(f"Contents integrity verified.")
        await sock.send(dumps({"call": "conf", "data": 1}))
        return True
    else:
        general(f"Contents integrity mismatch.")
        await sock.send(dumps({"call": "conf", "data": 0}))
        return False


async def collect_confirmation(data: int = 0) -> bool:
    if bool(data):
        general(f"Contents integrity verified.")
        return True
    else:
        general(f"Contents integrity mismatch.")
        return False



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


from asyncio import get_event_loop, ensure_future
from websockets import connect
from websockets.exceptions import ConnectionClosed
from tqdm.asyncio import tqdm

import sys
from expedite.config import standard
from expedite.client.conn import (
    deliver_connection_to_server, 
    collect_permission_to_join, 
    deliver_suspension_from_expiry, 
    collect_connection_from_pairness,
    deliver_dropping_summon,
    collect_dropping_summon,
    collect_metadata,
    deliver_metadata,
    collect_contents,
    deliver_contents,
    deliver_digest_checks,
    collect_digest_checks,
    deliver_confirmation,
    collect_confirmation,
    facade_exit
)
from json import loads


async def oper():
    try:
        async with connect(standard.client_addr) as sock:
            get_event_loop().call_later(standard.client_time, lambda: ensure_future(deliver_suspension_from_expiry(sock)))
            await deliver_connection_to_server(sock)
            async for mesgcont in sock:
                if isinstance(mesgcont, str):
                    mesgdict = loads(mesgcont)
                    if mesgdict["call"] == "okay":
                        await collect_permission_to_join(mesgdict["iden"])
                    elif mesgdict["call"] == "meta":
                        await collect_metadata(mesgdict["name"], mesgdict["size"], mesgdict["chks"])
                        if standard.client_plan == "RECV":
                            await deliver_dropping_summon(sock)
                    elif mesgdict["call"] == "note":
                        await collect_connection_from_pairness(mesgdict["part"])
                        if standard.client_plan == "SEND":
                            await deliver_metadata(sock)
                    elif mesgdict["call"] == "drop":
                        await collect_dropping_summon()
                        if standard.client_plan == "SEND":
                            await deliver_contents(sock)
                            await deliver_digest_checks(sock)
                    elif mesgdict["call"] == "byte":
                        if standard.client_plan == "RECV":
                            await collect_contents(mesgdict["pack"])
                    elif mesgdict["call"] == "hash":
                        if standard.client_plan == "RECV":
                            await collect_digest_checks()
                            complete = await deliver_confirmation(sock, mesgdict["data"])
                            await facade_exit(sock, complete, "done" if complete else "dprt")
                    elif mesgdict["call"] == "conf":
                        if standard.client_plan == "SEND":
                            complete = await collect_confirmation(mesgdict["data"])
                            await facade_exit(sock, complete, "done" if complete else "dprt")
                    elif mesgdict["call"] in ["awry", "lone"]:
                        await facade_exit(sock, False, mesgdict["call"])
                else:
                    if standard.client_plan == "RECV":
                        await collect_contents(sock, mesgcont)
        sys.exit(standard.client_exit)
    except ConnectionClosed as expt:
        await facade_exit(sock, False, "dprt")
        sys.exit(standard.client_exit)



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


from expedite.config import standard
from os.path import getsize, basename, exists


def find_size() -> int:
    return getsize(standard.client_file)


def find_name() -> str:
    return basename(standard.client_file)


def ease_size(size: int) -> str:
    unitlist = ["B", "KB", "MB", "GB", "TB", "PB"]
    indx, opsz = 0, size
    if size == 0:
        return "0 B"
    else:
        while opsz >= 1024 and indx < len(unitlist) - 1:
            opsz, indx = opsz / 1024.0, indx + 1
        return f"{opsz:.2f} {unitlist[indx]}"


def bite_file() -> list:
    init, size, bite = 0, standard.client_filesize, []
    while init < size:
        bite.append(init)
        if size - init >= standard.chunking_size:
            init = init + standard.chunking_size
        else:
            bite.append(size)
            init = size
    return bite


def read_file(init: int = 0, fina: int = 0) -> bytes:
    if exists(standard.client_file):
        with open(standard.client_file, "rb") as file:
            file.seek(init)
            data = file.read(fina - init)
            standard.client_hash.update(data)
        return data
    else:
        return b""


def fuse_file(pack: bytes = b"") -> bool:
    if not standard.client_fileinit:
        mode, standard.client_fileinit = "wb", True
    else:
        mode = "ab"
    with open(standard.client_filename, mode) as file:
        file.write(pack)
        standard.client_hash.update(pack)
    return True



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


import time

from expedite.view import warning, general, success, failure
from expedite.config import standard
from websockets.legacy.client import WebSocketClientProtocol


async def facade_exit(sock: WebSocketClientProtocol = None, cond: bool = True, note: str = "") -> None:
    if note != "done":
        warning(standard.notice_dict[note])
    if sock:
        await sock.close()
    plan = "Delivering" if standard.client_plan == "SEND" else "Collecting"
    if cond:
        success(f"{plan} done after {(time.time() - standard.client_strt):.2f} seconds.")
        standard.client_exit = 0
    else:
        failure(f"{plan} fail after {(time.time() - standard.client_strt):.2f} seconds.")
        standard.client_exit = 1
    general("Exiting.")



"""
expedite
Copyright (C) 2024 Akashdeep Dhar

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <https://www.gnu.org/licenses/>.

Any Red Hat trademarks that are incorporated in the codebase or documentation
are not subject to the GNU Affero General Public License and may only be used
or replicated with the express permission of Red Hat, Inc.
"""


from expedite.view import warning, success, failure, general
from expedite.config import standard
from uuid import uuid4
from json import dumps
from websockets.legacy.server import WebSocketServerProtocol


async def exchange_insert(
    sock: WebSocketServerProtocol,
    plan: str = standard.client_plan,
    scan: str = standard.client_endo,
    time: int = standard.client_time
) -> str | bool:
    if sock not in standard.connection_dict:
        if plan in ["SEND", "RECV"]:
            identity = uuid4().hex[0:8].upper()
            standard.connection_dict[sock] = {
                "iden": identity,
                "plan": plan,
                "scan": scan if scan != "" else None,
                "time": time,
                "pair": False,
                "ptid": "",
                "ptsc": None
            }
            if plan == "SEND":
                warning(f"{identity} joined with the intention of delivering.")
            elif plan == "RECV":
                warning(f"{identity} joined with the intention of collecting.")
            if scan == "":
                general(f"{identity} is waiting for client for {time} seconds.")
            else:
                general(f"{identity} is looking for {scan} for {time} seconds.")
            await sock.send(dumps({"call": "okay", "iden": identity}))
            return identity
        else:
            return False
    else:
        return False


async def exchange_remove(sock: WebSocketServerProtocol) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            warning(f"{standard.connection_dict[sock]["ptid"]} left.")
            await standard.connection_dict[sock]["ptsc"].close(code=1000)
            standard.connection_dict.pop(standard.connection_dict[sock]["ptsc"])
        warning(f"{standard.connection_dict[sock]["iden"]} left.")
        await sock.close(code=1000)
        standard.connection_dict.pop(sock)
        return True
    else:
        return False


async def exchange_inform(
    sock: WebSocketServerProtocol,
    plan: str = standard.client_plan,
    scan: str = standard.client_endo,
    iden: str = standard.client_iden
) -> int:
    for indx in standard.connection_dict:
        if standard.connection_dict[indx]["iden"] == scan:
            if not standard.connection_dict[indx]["pair"]:
                if plan != standard.connection_dict[indx]["plan"]:
                    success(f"{iden} and {scan} are positively paired.")
                    await indx.send(dumps({"call": "note", "part": iden}))
                    await sock.send(dumps({"call": "note", "part": scan}))
                    standard.connection_dict[indx]["pair"], standard.connection_dict[sock]["pair"] = True, True
                    standard.connection_dict[indx]["ptid"], standard.connection_dict[sock]["ptid"] = iden, scan
                    standard.connection_dict[indx]["ptid"], standard.connection_dict[sock]["ptid"] = iden, scan
                    standard.connection_dict[indx]["ptsc"], standard.connection_dict[sock]["ptsc"] = sock, indx
                    return 0
                else:
                    failure(f"{iden} and {scan} are negatively paired.")
                    await indx.send(dumps({"call": "awry", "part": iden}))
                    await sock.send(dumps({"call": "awry", "part": scan}))
                    return 1
            else:
                failure(f"{iden} and {scan} cannot pair as {scan} is already paired.")
                await sock.send(dumps({"call": "lone", "part": scan}))
                return 2
    return 3


async def exchange_launch(
    sock: WebSocketServerProtocol,
    name: str = standard.client_filename,
    size: str = standard.client_filesize,
    chks: str = standard.client_chks,
) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            general(f"{standard.connection_dict[sock]["iden"]} is attempting to share file metadata to {standard.connection_dict[sock]["ptid"]}.")
            await standard.connection_dict[sock]["ptsc"].send(dumps({"call": "meta", "name": name, "size": size, "chks": chks}))
            return True
        else:
            return False
    else:
        return False


async def exchange_gobyte(
    sock: WebSocketServerProtocol
) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            general(f"{standard.connection_dict[sock]["iden"]} is attempting to fetch file contents from {standard.connection_dict[sock]["ptid"]}.")
            await standard.connection_dict[sock]["ptsc"].send(dumps({"call": "drop"}))
            return True
        else:
            return False
    else:
        return False


async def exchange_detail(
    sock: WebSocketServerProtocol,
    pack: bytes = b""
) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            general(f"{standard.connection_dict[sock]["iden"]} is delivering contents to {standard.connection_dict[sock]["ptid"]}.")
            await standard.connection_dict[sock]["ptsc"].send(pack)
            return True
        else:
            general(f"{standard.connection_dict[sock]["iden"]} could not deliver contents to {standard.connection_dict[sock]["ptid"]} as {standard.connection_dict[sock]["ptid"]} is no longer connected.")
            return False
    else:
        general("Attempting for aborting connection.")
        return False


async def exchange_digest(
    sock: WebSocketServerProtocol,
    data: str = standard.client_hash.hexdigest()
) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            general(f"{standard.connection_dict[sock]["iden"]} is delivering digest to {standard.connection_dict[sock]["ptid"]}.")
            await standard.connection_dict[sock]["ptsc"].send(dumps({"call": "hash", "data": data}))
            return True
        else:
            return False
    else:
        return False


async def exchange_assert(
    sock: WebSocketServerProtocol,
    data: int = 0
) -> bool:
    if sock in standard.connection_dict:
        if standard.connection_dict[sock]["ptsc"] in standard.connection_dict and standard.connection_dict[sock]["ptsc"].state == 1:
            general(f"{standard.connection_dict[sock]["iden"]} is delivering confirmation to {standard.connection_dict[sock]["ptid"]}.")
            await standard.connection_dict[sock]["ptsc"].send(dumps({"call": "conf", "data": data}))
            return True
        else:
            return False
    else:
        return False
